import os
import logging
from dotenv import load_dotenv


def load_env():
    # Load comments from .env file
    env_path = os.path.join(os.path.dirname(__file__), ".env.public")
    if not env_path:
        assert False, f"🚫 api/app/settings: Could not find .env.public file"
    else:
        logging.info(f"✅ Loading secrets at {env_path}")
        load_dotenv(env_path)

    # Load specific environment variables from .env.{ENV} file
    ENV = os.getenv("ENV", "local")
    env_path = os.path.join(os.path.dirname(__file__), f".env.{ENV}")
    if not env_path:
        logging.info(f"🚫 api/app/settings:::.env.{ENV} file")
    else:
        logging.info(f"✅ Loading secrets at {env_path}")
        load_dotenv(env_path)


def getenv(key, default=None):
    value = os.getenv(key, default)
    if value == "True":
        return True
    elif value == "False":
        return False
    elif value == "None":
        return None
    else:
        return value



import clisync
import logging
from typing import Optional, List

from pyhectiqlab.const import DISABLE_BLOCK_CREATION
from pyhectiqlab.project import Project
from pyhectiqlab.client import Client
from pyhectiqlab.decorators import functional_alias, online_method


class Block:

    _block: Optional[str] = None

    @staticmethod
    @functional_alias("get_block")
    def get(block: Optional[str] = None):
        return block or Block._block

    @staticmethod
    @functional_alias("set_block")
    def set(block: str):
        Block._block = block

    @staticmethod
    def format_id(block: Optional[int] = None, project: Optional[str] = None):
        block = block or Block.get()
        project = project or Project.get()
        if not block or not project:
            return
        return f"{block}::{project.replace('/', '::')}"

    @staticmethod
    @online_method
    @functional_alias("create_block")
    @clisync.include(wait_response=True)
    def create(
        title: str, 
        status: Optional[str] = None, 
        project: Optional[str] = None,
        wait_response: Optional[bool] = False
    ):
        """Create a block.

        Args:
            title (str): Block title.
            status (str, optional): Block status. Default: None.
            project (str, optional): Project name. If None, the current project is used. Default: None.
        """
        if DISABLE_BLOCK_CREATION:
            logging.error("Block creation is disabled. Continuing anyway...")
            return
        project = project or Project.get()
        if not project:
            return None

        body = {"title": title, "status": status, "project": project}
        block = Client.post("/app/blocks", json=body, wait_response=wait_response)
        return block

    @staticmethod
    @online_method
    @functional_alias("update_block")
    @clisync.include(wait_response=True)
    def update(
        block: str, 
        title: Optional[str] = None, 
        status: Optional[str] = None, 
        wait_response: Optional[bool] = False
    ):
        """Update a block.

        Args:
            block (str): Block ID.
            title (str, optional): Block title. Default: None.
            status (str, optional): Block status. Default: None.
        """
        body = {"title": title, "status": status}
        # Remove None
        body = {k: v for k, v in body.items() if v is not None}
        if len(body) == 0:
            return
        block = Client.put(f"/app/blocks/{block}", json=body, wait_response=wait_response)
        return block

    @staticmethod
    @online_method
    @functional_alias("list_blocks")
    @clisync.include()
    def list(
        project: Optional[str] = None,
        search: Optional[str] = None,
        fields: Optional[List[str]] = None,
        page: Optional[int] = 1,
        limit: Optional[int] = 50,
        order_by: Optional[str] = "created_at",
        order_direction: Optional[str] = "desc",
    ):
        """List blocks.

        Args:
            project (str, optional): Project name. If None, the current project is used. Default: None.
            search (str, optional): Search string. Default: None.
        """
        project = project or Project.get()
        if not project:
            return None
        params = {
            "project": project,
            "search": search,
            "fields": fields,
            "page": page,
            "limit": limit,
            "order_by": order_by,
            "order_direction": order_direction,
        }
        blocks = Client.get("/app/blocks", params=params, wait_response=True)
        return blocks



from collections import defaultdict
from typing import Callable, List, Literal, Optional
from functools import partial

import threading
import numpy as np
import time
import datetime as dt
import logging

logger = logging.getLogger("pyhectiqlab")


class RepeatedTimer(threading.Thread):
    """
    RepeatedTimer class is used to run a method at a given interval.

    Args:
        method (callable): Method to run.
        delay (int): Delay between two executions.
    """

    def __init__(self, method: Callable[[], None], delay: int):
        super().__init__()
        self.daemon = True
        self._stopper = threading.Event()
        self.delay = delay
        self.method = method

    def run(self):
        while not self._stopper.wait(self.delay):
            self.method()

    def stopped(self):
        return self._stopper.is_set()

    def stop(self):
        self._stopper.set()


class MetricsManager:
    """
    MetricsManager class is used to manage the metrics cache and push data to the server.

    Args:
        push_method (callable, optional): Function to push data to the server
        max_cache_timeout (int): Maximum number of seconds in the cache
        max_cache_length (int): Number of elements in the cache to trigger a cache dump
        min_cache_flush_delay (int): Minimum number of seconds between two dumps
        aggregate_metrics (str): Wheter to push all data in cache or aggregate
    """

    __max_cache_quantity_per_second = 200

    def __init__(
        self,
        push_method: Callable[[str, List[float]], None],
        run_id: Optional[str] = None,
        max_cache_timeout: int = 10,
        max_cache_length: int = 100,
        min_cache_flush_delay: int = 5,
        aggregate_metrics: str = "none",
    ):
        self.cache = defaultdict(list)
        self.cache_last_push = {}
        self._warning_is_shown = False
        self.max_cache_length = max_cache_length
        self.min_cache_flush_delay = min_cache_flush_delay

        self.push_method = partial(push_method, run_id=run_id)
        self.set_aggr(aggregate_metrics)

        self.timer = RepeatedTimer(self.flush_cache, max_cache_timeout)
        self.timer.start()
        return

    def update_cache_settings(
        self,
        max_cache_timeout: Optional[int] = None,
        max_cache_length: Optional[int] = None,
        min_cache_flush_delay: Optional[int] = None,
    ):
        """
        Update cache settings.
        """

        if max_cache_timeout is not None:
            self.max_cache_timeout = max_cache_timeout
            self.timer.stop()
            self.timer = RepeatedTimer(self.flush_cache, max_cache_timeout)
            self.timer.start()
        if max_cache_length is not None:
            self.max_cache_length = max_cache_length
        if min_cache_flush_delay is not None:
            self.min_cache_flush_delay = min_cache_flush_delay

    def set_aggr(self, new_value: Literal["none", "sum", "max", "mean"]):
        vals = ["none", "sum", "max", "mean"]
        assert new_value in vals, f"Aggr must be in {vals}"
        self.aggregate_metrics = new_value

    def add(self, key: str, step: int, value: float) -> None:
        """
        Add a new metric to the cache.
        """
        if value is None:
            return
        self.cache[key].append((float(step), float(value), dt.datetime.now(tz=dt.timezone.utc).timestamp()))
        if len(self.cache[key]) > self.max_cache_length:
            if key in self.cache_last_push:
                last_push_time = self.cache_last_push[key]
                elapsed = time.time() - last_push_time
                if elapsed < self.min_cache_flush_delay:
                    return
            self.push_data(key)

    def subsample(self, data: List[float]) -> List[float]:
        normalized_max = int(self.__max_cache_quantity_per_second * self.min_cache_flush_delay)
        if len(data) <= normalized_max:
            return data
        if self._warning_is_shown == False:
            logger.warning(
                f"Your metrics rate ({len(data)//self.min_cache_flush_delay} metrics/second) is exceeding the "
                "maximum rate ({MetricsManager.__max_cache_quantity_per_second}/second). Your metrics have been "
                "subsampled automatically."
            )
            logger.warning(
                "Consider reducing the number of metrics you push or setting an aggregation method with "
                "`run.set_metrics_aggr('mean')`."
            )
            logger.warning("See the documentation for details https://docs.hectiq.ai/objects/metrics/.")
            self._warning_is_shown = True
        indexes = np.linspace(0, len(data) - 1, normalized_max).astype(int)
        return np.array(data)[indexes].tolist()

    def push_data(self, key: str) -> None:
        self.cache_last_push[key] = time.time()
        if self.push_method is None:
            self.cache[key] = []
            if self._warning_is_shown == False:
                logger.warning(
                    "No push method has been set. Please set a push method with `run.set_metrics_push_method`."
                )
            return

        if self.aggregate_metrics == "none":
            data = self.subsample(self.cache[key])
            self.push_method(key, data)
        elif self.aggregate_metrics == "mean":
            k = [_[0] for _ in self.cache[key]]
            s = [_[1] for _ in self.cache[key]]
            t = [_[2] for _ in self.cache[key]]
            if len(k) > 0:
                self.push_method(key, [(float(np.max(k)), float(np.mean(s)), float(np.mean(s)))])
        elif self.aggregate_metrics == "max":
            k = [_[0] for _ in self.cache[key]]
            s = [_[1] for _ in self.cache[key]]
            t = [_[2] for _ in self.cache[key]]
            if len(k) > 0:
                self.push_method(key, [(float(np.max(k)), float(np.max(s)), float(np.mean(s)))])
        elif self.aggregate_metrics == "sum":
            k = [_[0] for _ in self.cache[key]]
            s = [_[1] for _ in self.cache[key]]
            t = [_[2] for _ in self.cache[key]]
            if len(k) > 0:
                self.push_method(key, [(float(np.max(k)), float(np.sum(s)), float(np.mean(t)))])
        self.cache[key] = []

    def __delete__(self):
        self.timer.stop()
        self.flush_cache()

    def flush_cache(self):
        for key in self.cache:
            self.push_data(key)



__version__ = "3.0.8"

from pyhectiqlab.settings import getenv

API_URL = getenv("HECTIQLAB_API_URL", "https://api.lab.hectiq.ai")

from pyhectiqlab.config import Config
from pyhectiqlab.dataset import Dataset
from pyhectiqlab.model import Model
from pyhectiqlab.run import Run
from pyhectiqlab import functional

from pyhectiqlab.logging import setup_logging
setup_logging()

def debug_mode():
    from pyhectiqlab.client import Client
    Client.online(False)

__all__ = [
    "API_URL",
    "Artifact",
    "Dataset",
    "Model",
    "Run",
    "Config",
    "Block",
    "functional",
    "debug_mode",
]



import pytest

from pyhectiqlab import Run
from mock_client import mock_client
from utils import random_uuid


@pytest.fixture
def config():
    return {"param": "test", "x": 1}


def test_run_create():
    with mock_client(Run) as r:
        run = r(title="test run", project="hectiq-ai/test")


def test_run_retrieve():
    run = Run(rank=1, project="hectiq-ai/test")


@pytest.mark.filterwarnings("error")
def test_run_retrieve_by_id():

    # retrieving first run by id
    run_id = "7abdd377-b702-452c-bfe5-df13f34a99d9"
    run = Run.retrieve_by_id(run_id=run_id)
    assert run is not None
    assert run["id"] == run_id


@pytest.mark.filterwarnings("error")
def test_run_retrieve_not_existing_rank_raises_warning():
    invalid_rank = 1000
    with pytest.raises(UserWarning):
        run = Run(rank=invalid_rank, project="hectiq-ai/test")


def test_run_exists():
    assert Run.exists(rank=1, project="hectiq-ai/test") == True
    assert Run.exists(rank=1000, project="hectiq-ai/test") == False


def test_run_add_config(config):
    run = Run(rank=1, project="hectiq-ai/test")
    config = {"param": "test", "x": 1}
    run.add_config(config)


def test_run_retrieve_config(config):
    run = Run(rank=1, project="hectiq-ai/test")
    assert run.retrieve_config() == config


def test_run_add_artifact():
    import os

    with mock_client(Run) as r:
        run = r(rank=1, project="hectiq-ai/test")
        path = os.path.join(os.path.dirname(__file__), "dummy/artifact.txt")
        res = run.add_artifact(path, name="content.txt", wait_response=True)

def test_run_add_figure():
    run = Run(title=random_uuid(), project="hectiq-ai/test")
    import matplotlib.pyplot as plt
    plt.style.use("dark_background")
    plt.figure(figsize=(3, 3))
    plt.plot([1, 2, 3, 4], [1, 4, 9, 16])
    run.add_figure(plt, name="test-figure", wait_response=True)

def test_run_set_status():
    run = Run(rank=1, project="hectiq-ai/test")
    run.set_status("test status")


def test_run_add_tag():
    run = Run(rank=1, project="hectiq-ai/test")
    run.add_tags("some")


def test_run_detach_tag():
    run = Run(rank=1, project="hectiq-ai/test")
    run.add_tags("some")
    run.detach_tag("some")


def test_run_attach_dataset():
    run = Run(rank=1, project="hectiq-ai/test")
    print(run.attach_dataset(name="test-dataset", version="1.1", wait_response=True))


def test_run_attach_model():
    run = Run(rank=1, project="hectiq-ai/test")
    print(run.attach_model(name="test-model", version="1.0", wait_response=True))


if __name__ == "__main__":
    pass



import platform
from multiprocessing import cpu_count
import socket
import importlib
from typing import Optional, List
import clisync
import inspect
import types
import sys
from pyhectiqlab.project import Project
from pyhectiqlab.decorators import functional_alias
from pyhectiqlab import settings

import logging
logger = logging.getLogger()

from pyhectiqlab.logging import hectiqlab_logger

try:
    from git import Repo
except ImportError:
    pass

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata


class PackageVersion:

    @staticmethod
    @functional_alias("get_package_version")
    @clisync.include()
    def all(repos: Optional[List[str]] = None) -> dict:
        """Get all packages and system information."""
        packages = {}
        try:
            packages["python"] = PackageVersion.python_version()
        except Exception as e:
            logger.error(f"An error occurred while getting Python version: {e}")
        try:
            packages["system"] = PackageVersion.sysinfo()
        except Exception as e:
            logger.error(f"An error occurred while getting system information: {e}")
        try:
            packages["packages"] = PackageVersion.imported_modules_versions()
        except Exception as e:
            logger.error(f"An error occurred while getting package versions: {e}")
        try:
            repos = repos or []
            default_repos = settings.getenv("HECTIQ_LAB_REPOS", None)
            if default_repos:
                repos = default_repos.split(",") + repos

            if repos:
                packages["repos"] = {}
                for repo in repos:
                    packages["repos"][repo] = PackageVersion.repo(repo)
        except Exception as e:
            logger.error(f"An error occurred while getting git repository information: {e}")
        hectiqlab_logger.debug(f"Package versions: {packages}")
        return packages

    @staticmethod
    def python_version():
        """Get Python version."""
        return {"Python implementation": platform.python_implementation(), "Python version": platform.python_version()}

    @staticmethod
    def sysinfo():
        """Get system information."""
        return {
            "Compiler": platform.python_compiler(),
            "OS": platform.system(),
            "Release": platform.release(),
            "Machine": platform.machine(),
            "Processor": platform.processor(),
            "CPU cores": cpu_count(),
            "Architecture": platform.architecture()[0],
            "Hostname": socket.gethostname(),
        }

    @staticmethod
    def repo(package_name: str):
        """Get git repository information of a package."""
        repo = PackageVersion._get_repo(package_name)
        if repo is None:
            return
        diff_files = None
        try:
            has_diff = repo.is_dirty()
            if has_diff:
                diff_files = repo.git.diff("--name-only").split("\n")
        except:
            has_diff = "Unknown"

        return {
            "branch": repo.active_branch.name,
            "commit": repo.active_branch.commit.hexsha,
            "remote": repo.remotes.origin.url,
            "has_diff": has_diff,
            "diff_files": diff_files,
        }

    @staticmethod
    def is_dirty(package_name: Optional[str] = None) -> Optional[bool]:
        """Indicates if the package is dirty."""
        package_name = package_name or Project.repo
        if package_name is None:
            return False
        repo = PackageVersion._get_repo(package_name)
        if repo is None:
            return
        try:
            return repo.is_dirty()
        except:
            return None

    @staticmethod
    def _get_repo(package_name: str):
        """Get the git repository of a package.

        Args:
            package_name (str): Name of the package.
        """
        try:
            spec = importlib.util.find_spec(package_name)
        except:
            return
        if spec is None:
            return

        path = spec.origin
        try:
            repo = Repo(path, search_parent_directories=True)
        except:
            return

        return repo

    @staticmethod
    def imported_modules_versions():
        global_namespaces = sys.modules
        to_print = {}
        modules = {
            val.__name__.split(".")[0] for _, val in global_namespaces.items() if isinstance(val, types.ModuleType)
        }
        submodules = {
            inspect.getmodule(val).__name__.split(".")[0]
            for _, val in global_namespaces.items()
            if (inspect.getmodule(val) is not None)
        }
        modules = set.union(submodules, modules)

        modules.discard("builtins")
        modules.discard("IPython")
        modules.discard("__main__")

        for module in modules:
            if module.startswith("_"):
                continue
            pkg_version = PackageVersion.module_version(module)
            if pkg_version not in ("not installed", "unknown"):
                to_print[module] = pkg_version

        # Sort keys
        to_print = dict(sorted(to_print.items()))
        return to_print

    @staticmethod
    def module_version(module):
        """Return the version of a given package"""
        if module == "scikit-learn":
            module = "sklearn"
        try:
            imported = importlib.import_module(module)
        except ImportError:
            version = "not installed"
        else:
            try:
                version = importlib_metadata.version(module)
            except importlib_metadata.PackageNotFoundError:
                try:
                    version = imported.__version__
                except AttributeError:
                    try:
                        version = imported.version
                    except AttributeError:
                        try:
                            version = imported.version_info
                        except AttributeError:
                            version = "unknown"
        return version
